import datetime as dt

import attr
from dateutil.parser import isoparse as parse_iso_date

from .base import BaseCZMLObject
from .constants import ISO8601_FORMAT_Z

TYPE_MAPPING = {bool: "boolean"}


def format_datetime_like(dt_object):
    if dt_object is None:
        result = dt_object

    elif isinstance(dt_object, str):
        try:
            parse_iso_date(dt_object)
        except Exception:
            raise
        else:
            result = dt_object

    elif isinstance(dt_object, dt.datetime):
        result = dt_object.astimezone(dt.timezone.utc).strftime(ISO8601_FORMAT_Z)

    else:
        result = dt_object.strftime(ISO8601_FORMAT_Z)

    return result


@attr.s(str=False, frozen=True, kw_only=True)
class _TimeTaggedCoords(BaseCZMLObject):

    NUM_COORDS: int

    values = attr.ib()

    @values.validator
    def _check_values(self, attribute, value):
        if not (
            len(value) == self.NUM_COORDS or len(value) % (self.NUM_COORDS + 1) == 0
        ):
            raise ValueError(
                "Input values must have either 3 or N * 4 values, "
                "where N is the number of time-tagged samples."
            )

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class FontValue(BaseCZMLObject):
    """A font, specified using the same syntax as the CSS "font" property."""

    font = attr.ib(default=None)

    def to_json(self):
        return self.font


@attr.s(str=False, frozen=True, kw_only=True)
class RgbafValue(BaseCZMLObject):
    """A color specified as an array of color components [Red, Green, Blue, Alpha]
     where each component is in the range 0.0-1.0. If the array has four elements,
    the color is constant. If it has five or more elements, they are time-tagged
    samples arranged as [Time, Red, Green, Blue, Alpha, Time, Red, Green, Blue, Alpha, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.

    """

    values = attr.ib()

    @values.validator
    def _check_values(self, attribute, value):
        if not (len(value) == 4 or len(value) % 5 == 0):
            raise ValueError(
                "Input values must have either 4 or N * 5 values, "
                "where N is the number of time-tagged samples."
            )

        if len(value) == 4:
            if not all([0 <= val <= 1 for val in value]):
                raise ValueError("Color values must be floats in the range 0-1.")

        else:
            for i in range(0, len(value), 5):
                v = value[i + 1 : i + 5]

                if not all([0 <= val <= 1 for val in v]):
                    raise ValueError("Color values must be floats in the range 0-1.")

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class RgbaValue(BaseCZMLObject):
    """A color specified as an array of color components [Red, Green, Blue, Alpha]
    where each component is in the range 0-255. If the array has four elements,
    the color is constant.

    If it has five or more elements, they are time-tagged samples arranged as
    [Time, Red, Green, Blue, Alpha, Time, Red, Green, Blue, Alpha, ...], where Time
    is an ISO 8601 date and time string or seconds since epoch.

    """

    values = attr.ib()

    @values.validator
    def _check_values(self, attribute, value):
        if not (len(value) == 4 or len(value) % 5 == 0):
            raise ValueError(
                "Input values must have either 4 or N * 5 values, "
                "where N is the number of time-tagged samples."
            )

        if len(value) == 4:
            if not all([type(val) is int and 0 <= val <= 255 for val in value]):
                raise ValueError("Color values must be integers in the range 0-255.")

        else:
            for i in range(0, len(value), 5):
                v = value[i + 1 : i + 5]

                if not all([type(val) is int and 0 <= val <= 255 for val in v]):
                    raise ValueError(
                        "Color values must be integers in the range 0-255."
                    )

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class ReferenceValue(BaseCZMLObject):
    """Represents a reference to another property. References can be used to specify that two properties on different
    objects are in fact, the same property.

    """

    string = attr.ib(default=None)

    @string.validator
    def _check_string(self, attribute, value):
        if not isinstance(value, str):
            raise ValueError("Reference must be a string")
        if "#" not in value:
            raise ValueError(
                "Invalid reference string format. Input must be of the form id#property"
            )

    def to_json(self):
        return self.string


@attr.s(str=False, frozen=True, kw_only=True)
class Cartesian3Value(_TimeTaggedCoords):
    """A three-dimensional Cartesian value specified as [X, Y, Z].

    If the values has three elements, the value is constant.
    If it has four or more elements, they are time-tagged samples
    arranged as [Time, X, Y, Z, Time, X, Y, Z, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.

    """

    NUM_COORDS = 3


@attr.s(str=False, frozen=True, kw_only=True)
class CartographicRadiansValue(_TimeTaggedCoords):
    """A geodetic, WGS84 position specified as [Longitude, Latitude, Height].

    Longitude and Latitude are in radians and Height is in meters.
    If the array has three elements, the value is constant.
    If it has four or more elements, they are time-tagged samples
    arranged as [Time, Longitude, Latitude, Height, Time, Longitude, Latitude, Height, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.

    """

    NUM_COORDS = 3


@attr.s(str=False, frozen=True, kw_only=True)
class CartographicDegreesValue(_TimeTaggedCoords):
    """A geodetic, WGS84 position specified as [Longitude, Latitude, Height].

    Longitude and Latitude are in degrees and Height is in meters.
    If the array has three elements, the value is constant.
    If it has four or more elements, they are time-tagged samples
    arranged as [Time, Longitude, Latitude, Height, Time, Longitude, Latitude, Height, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.

    """

    NUM_COORDS = 3


@attr.s(str=False, frozen=True, kw_only=True)
class StringValue(BaseCZMLObject):
    """A string value.

    The string can optionally vary with time.
    """

    string = attr.ib(default=None)

    def to_json(self):
        return self.string


@attr.s(str=False, frozen=True, kw_only=True)
class CartographicRadiansListValue(BaseCZMLObject):
    """A list of geodetic, WGS84 positions specified as [Longitude, Latitude, Height, Longitude, Latitude, Height, ...],
    where Longitude and Latitude are in radians and Height is in meters."""

    values = attr.ib()

    @values.validator
    def _check_values(self, attribute, value):
        if len(value) % 3 != 0:
            raise ValueError(
                "Invalid values. Input values should be arrays of size 3 * N"
            )

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class CartographicDegreesListValue(BaseCZMLObject):
    """A list of geodetic, WGS84 positions specified as [Longitude, Latitude, Height, Longitude, Latitude, Height, ...],
    where Longitude and Latitude are in degrees and Height is in meters."""

    values = attr.ib()

    @values.validator
    def _check_values(self, attribute, value):
        if len(value) % 3 != 0:
            raise ValueError(
                "Invalid values. Input values should be arrays of size 3 * N"
            )

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class DistanceDisplayConditionValue(BaseCZMLObject):
    """A value indicating the visibility of an object based on the distance to the camera, specified as two values
    [NearDistance, FarDistance]. If the array has two elements, the value is constant. If it has three or more elements,
    they are time-tagged samples arranged as [Time, NearDistance, FarDistance, Time, NearDistance, FarDistance, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.
    """

    values = attr.ib(default=None)

    @values.validator
    def _check_values(self, attribute, value):
        if len(value) != 2 and len(value) % 3 != 0:
            raise ValueError(
                "Invalid values. Input values should be arrays of size either 2 or 3 * N"
            )

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class NearFarScalarValue(BaseCZMLObject):
    """A near-far scalar value specified as four values [NearDistance, NearValue, FarDistance, FarValue].

     If the array has four elements, the value is constant. If it has five or more elements, they are time-tagged
    samples arranged as [Time, NearDistance, NearValue, FarDistance, FarValue, Time, NearDistance, NearValue,
    FarDistance, FarValue, ...], where Time is an ISO 8601 date and time string or seconds since epoch.
    """

    values = attr.ib(default=None)

    @values.validator
    def _check_values(self, attribute, value):
        if not (len(value) == 4 or len(value) % 5 == 0):
            raise ValueError(
                "Input values must have either 4 or N * 5 values, "
                "where N is the number of time-tagged samples."
            )

    def to_json(self):
        return list(self.values)


@attr.s(str=False, frozen=True, kw_only=True)
class TimeInterval(BaseCZMLObject):
    """A time interval, specified in ISO8601 interval format."""

    _start = attr.ib(default=None)
    _end = attr.ib(default=None)

    def to_json(self):
        if self._start is None:
            start = "0000-00-00T00:00:00Z"
        else:
            start = format_datetime_like(self._start)

        if self._end is None:
            end = "9999-12-31T24:00:00Z"
        else:
            end = format_datetime_like(self._end)

        return "{start}/{end}".format(start=start, end=end)


@attr.s(str=False, frozen=True, kw_only=True)
class IntervalValue(BaseCZMLObject):
    """Value over some interval."""

    _start = attr.ib()
    _end = attr.ib()
    _value = attr.ib()

    def to_json(self):
        obj_dict = {"interval": TimeInterval(start=self._start, end=self._end)}

        try:
            obj_dict.update(**self._value.to_json())
        except AttributeError:
            key = TYPE_MAPPING[type(self._value)]
            obj_dict[key] = self._value

        return obj_dict


@attr.s(str=False, frozen=True)
class Sequence(BaseCZMLObject):
    """Sequence, list, array of objects."""

    _values = attr.ib()

    def to_json(self):
        return list(self._values)


@attr.s(str=False, frozen=True, kw_only=True)
class UnitQuaternionValue(_TimeTaggedCoords):
    """A set of 4-dimensional coordinates used to represent rotation in 3-dimensional space.

    It's specified as [X, Y, Z, W]. If the array has four elements, the value is constant.
    If it has five or more elements, they are time-tagged samples arranged as
    [Time, X, Y, Z, W, Time, X, Y, Z, W, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.

    """

    NUM_COORDS = 4



from functools import reduce

from .properties import Color
from .types import RgbafValue, RgbaValue


def get_color_list(timestamps, colors, rgbaf=False):
    """
    Given a list of valid colors (rgb/rgba tuples, shorthand hex string or integer representation) and a list of
    time-stamps, create a Color object with rgba/rgbaf values of the form: [time_0, r_0, g_0, b_0,..., time_k, r_k,
    g_k, b_k]

    Parameters
    ----------
    :param list(str) timestamps: The list of the timestamps (ISO 8601 date or seconds since epoch)
    :param str/int/list colors : A list of valid colors
    :param bool rgbaf: If set to True, returns rgbaf values, else return rgba values
    """
    # Check if colors is a valid list of colors
    if all(Color.is_valid(v) for v in colors):
        color_lst = list(map(get_color, colors))
    else:
        raise ValueError("Invalid input")

    # Quick function to convert between rgba-rgbaf easier
    if rgbaf:
        color_r = (
            lambda c: c.rgbaf.values
            if c.rgbaf
            else list(map(lambda x: float(x / 255), c.rgba.values))
        )
    else:
        color_r = (
            lambda c: c.rgba.values
            if c.rgba
            else list(map(lambda x: round(x * 255), c.rgbaf.values))
        )

    # Get combined list of timestamps and colors
    time_colr = [[time] + color_r(c) for time, c in zip(timestamps, color_lst)]
    # Flatten list
    time_colr = reduce(lambda x, y: x + y, time_colr)

    if rgbaf:
        return Color(rgbaf=RgbafValue(values=time_colr))
    else:
        return Color(rgba=RgbaValue(values=time_colr))


def get_color(color):
    """
    A helper function to make color setting more versatile. What the ``color`` parameter determines depends on
    its type.

    Parameters
    ----------

    :param  str/int/list color: Depending on the type, ``color`` can be either a hexadecimal rgb/rgba color value or
        a tuple in the form of [r, g, b, a] or [r, g, b].

    """
    # Color.from_string, Color.from_int, ...
    if isinstance(color, str) and 6 <= len(color) <= 10:
        return Color.from_str(color)
    elif issubclass(int, type(color)):
        return Color.from_hex(int(color))
    elif isinstance(color, list):
        # If it is a valid color in list form, simply return it
        if Color.is_valid(color):
            return Color.from_list(color)
    raise ValueError("Invalid input")



import datetime as dt

import pytest

from czmlpy.enums import ArcTypes, ClassificationTypes, ShadowModes
from czmlpy.properties import (
    ArcType,
    Box,
    BoxDimensions,
    CheckerboardMaterial,
    ClassificationType,
    Color,
    DistanceDisplayCondition,
    Ellipsoid,
    EllipsoidRadii,
    EyeOffset,
    GridMaterial,
    ImageMaterial,
    Material,
    Model,
    NearFarScalar,
    Orientation,
    Point,
    Polyline,
    PolylineArrow,
    PolylineArrowMaterial,
    PolylineDash,
    PolylineDashMaterial,
    PolylineGlow,
    PolylineGlowMaterial,
    PolylineMaterial,
    PolylineOutline,
    PolylineOutlineMaterial,
    Position,
    PositionList,
    ShadowMode,
    SolidColorMaterial,
    StripeMaterial,
    Uri,
    ViewFrom,
)
from czmlpy.types import (
    Cartesian3Value,
    CartographicDegreesListValue,
    DistanceDisplayConditionValue,
    IntervalValue,
    NearFarScalarValue,
    Sequence,
    UnitQuaternionValue,
)


def test_box():
    expected_result = """{
    "show": true,
    "dimensions": {
        "cartesian": [
            5,
            6,
            3
        ]
    }
}"""

    box = Box(
        show=True, dimensions=BoxDimensions(cartesian=Cartesian3Value(values=[5, 6, 3]))
    )
    assert str(box) == expected_result


def test_eyeOffset():
    expected_result = """{
    "cartesian": [
        1,
        2,
        3
    ]
}"""

    eyeOffset = EyeOffset(cartesian=Cartesian3Value(values=[1, 2, 3]))
    assert str(eyeOffset) == expected_result


def test_point():
    expected_result = """{
    "show": true,
    "pixelSize": 10,
    "scaleByDistance": {
        "nearFarScalar": [
            150,
            2.0,
            15000000,
            0.5
        ]
    },
    "disableDepthTestDistance": 1.2
}"""

    pnt = Point(
        show=True,
        pixelSize=10,
        scaleByDistance=NearFarScalar(
            nearFarScalar=NearFarScalarValue(values=[150, 2.0, 15000000, 0.5])
        ),
        disableDepthTestDistance=1.2,
    )
    assert str(pnt) == expected_result


def test_arc_type():
    expected_result = """{
    "arcType": "NONE"
}"""
    arc_type = ArcType(arcType=ArcTypes.NONE)
    assert str(arc_type) == expected_result


def test_shadow_mode():
    expected_result = """{
    "shadowMode": "ENABLED"
}"""
    shadow_mode = ShadowMode(shadowMode=ShadowModes.ENABLED)
    assert str(shadow_mode) == expected_result


def test_polyline():
    expected_result = """{
    "positions": {
        "cartographicDegrees": [
            20,
            30,
            10
        ]
    },
    "arcType": {
        "arcType": "GEODESIC"
    },
    "distanceDisplayCondition": {
        "distanceDisplayCondition": [
            14,
            81
        ]
    },
    "classificationType": {
        "classificationType": "CESIUM_3D_TILE"
    }
}"""
    pol = Polyline(
        positions=PositionList(
            cartographicDegrees=CartographicDegreesListValue(values=[20, 30, 10])
        ),
        arcType=ArcType(arcType="GEODESIC"),
        distanceDisplayCondition=DistanceDisplayCondition(
            distanceDisplayCondition=DistanceDisplayConditionValue(values=[14, 81])
        ),
        classificationType=ClassificationType(
            classificationType=ClassificationTypes.CESIUM_3D_TILE
        ),
    )
    assert str(pol) == expected_result


def test_material_solid_color():
    expected_result = """{
    "solidColor": {
        "color": {
            "rgba": [
                200,
                100,
                30,
                255
            ]
        }
    }
}"""
    mat = Material(solidColor=SolidColorMaterial.from_list([200, 100, 30]))

    assert str(mat) == expected_result

    pol_mat = PolylineMaterial(solidColor=SolidColorMaterial.from_list([200, 100, 30]))
    assert str(pol_mat) == expected_result


def test_arrowmaterial_color():
    expected_result = """{
    "polylineArrow": {
        "color": {
            "rgba": [
                200,
                100,
                30,
                255
            ]
        }
    }
}"""
    pamat = PolylineArrowMaterial(
        polylineArrow=PolylineArrow(color=Color(rgba=[200, 100, 30, 255])),
    )

    assert str(pamat) == expected_result


def test_dashmaterial_colors():
    expected_result = """{
    "polylineDash": {
        "color": {
            "rgba": [
                200,
                100,
                30,
                255
            ]
        },
        "gapColor": {
            "rgba": [
                100,
                200,
                0,
                255
            ]
        },
        "dashLength": 16,
        "dashPattern": 255
    }
}"""
    dashmat = PolylineDashMaterial(
        polylineDash=PolylineDash(
            color=Color(rgba=[200, 100, 30, 255]),
            gapColor=Color(rgba=[100, 200, 0, 255]),
            dashLength=16,
            dashPattern=255,
        ),
    )

    assert str(dashmat) == expected_result


def test_glowmaterial_color():
    expected_result = """{
    "polylineGlow": {
        "color": {
            "rgba": [
                200,
                100,
                30,
                255
            ]
        },
        "glowPower": 0.7,
        "taperPower": 0.3
    }
}"""
    glowmat = PolylineGlowMaterial(
        polylineGlow=PolylineGlow(
            color=Color(rgba=[200, 100, 30, 255]), glowPower=0.7, taperPower=0.3
        )
    )
    assert str(glowmat) == expected_result


def test_outline_material_colors():
    expected_result = """{
    "polylineOutline": {
        "color": {
            "rgba": [
                200,
                100,
                30,
                255
            ]
        },
        "outlineColor": {
            "rgba": [
                100,
                200,
                0,
                255
            ]
        },
        "outlineWidth": 3
    }
}"""
    omat = PolylineOutlineMaterial(
        polylineOutline=PolylineOutline(
            color=Color(rgba=[200, 100, 30, 255]),
            outlineColor=Color(rgba=[100, 200, 0, 255]),
            outlineWidth=3,
        )
    )
    assert str(omat) == expected_result


def test_color_isvalid():
    assert Color.is_valid([255, 204, 0, 55])
    assert Color.is_valid([255, 204, 55])
    assert Color.is_valid(0xFF3223)
    assert Color.is_valid(32)
    assert Color.is_valid(0xFF322332)
    assert Color.is_valid("#FF3223")
    assert Color.is_valid("#FF322332")
    assert Color.is_valid((255, 204, 55))
    assert Color.is_valid((255, 204, 55, 255))
    assert Color.is_valid((0.127568, 0.566949, 0.550556))
    assert Color.is_valid((0.127568, 0.566949, 0.550556, 1.0))


def test_color_isvalid_false():
    assert Color.is_valid([256, 204, 0, 55]) is False
    assert Color.is_valid([-204, 0, 55]) is False
    assert Color.is_valid([249.1, 204.3, 55.4]) is False
    assert Color.is_valid([255, 204]) is False
    assert Color.is_valid([255, 232, 300]) is False
    assert Color.is_valid(0xFF3223324) is False
    assert Color.is_valid(-3) is False
    assert Color.is_valid("totally valid color") is False
    assert Color.is_valid("#FF322332432") is False
    assert Color.is_valid((255, 204, 55, 255, 42)) is False
    assert Color.is_valid((0.127568, 0.566949, 0.550556, 1.0, 3.0)) is False


def test_material_image():
    expected_result = """{
    "image": {
        "image": "https://site.com/image.png",
        "repeat": [
            2,
            2
        ],
        "color": {
            "rgba": [
                200,
                100,
                30,
                255
            ]
        },
        "transparent": false
    }
}"""

    mat = Material(
        image=ImageMaterial(
            image=Uri(uri="https://site.com/image.png"),
            repeat=[2, 2],
            color=Color.from_list([200, 100, 30]),
        )
    )
    assert str(mat) == expected_result

    pol_mat = PolylineMaterial(
        image=ImageMaterial(
            image=Uri(uri="https://site.com/image.png"),
            repeat=[2, 2],
            color=Color.from_list([200, 100, 30]),
        )
    )
    assert str(pol_mat) == expected_result


def test_material_grid():
    expected_result = """{
    "color": {
        "rgba": [
            20,
            20,
            30,
            255
        ]
    },
    "cellAlpha": 1.0,
    "lineCount": [
        16,
        16
    ],
    "lineThickness": [
        2.0,
        2.0
    ],
    "lineOffset": [
        0.3,
        0.4
    ]
}"""

    pol_mat = GridMaterial(
        color=Color.from_list([20, 20, 30]),
        cellAlpha=1.0,
        lineCount=[16, 16],
        lineThickness=[2.0, 2.0],
        lineOffset=[0.3, 0.4],
    )
    assert str(pol_mat) == expected_result


def test_material_stripe():
    expected_result = """{
    "orientation": "HORIZONTAL",
    "evenColor": {
        "rgba": [
            0,
            0,
            0,
            255
        ]
    },
    "oddColor": {
        "rgba": [
            255,
            255,
            255,
            255
        ]
    },
    "offset": 0.3,
    "repeat": 4
}"""

    pol_mat = StripeMaterial(
        evenColor=Color.from_list([0, 0, 0]),
        oddColor=Color.from_list([255, 255, 255]),
        offset=0.3,
        repeat=4,
    )
    assert str(pol_mat) == expected_result


def test_material_checkerboard():
    expected_result = """{
    "evenColor": {
        "rgba": [
            0,
            0,
            0,
            255
        ]
    },
    "oddColor": {
        "rgba": [
            255,
            255,
            255,
            255
        ]
    },
    "repeat": 4
}"""

    pol_mat = CheckerboardMaterial(
        evenColor=Color.from_list([0, 0, 0]),
        oddColor=Color.from_list([255, 255, 255]),
        repeat=4,
    )
    assert str(pol_mat) == expected_result


def test_position_has_delete():
    pos = Position(delete=True, cartesian=[])

    assert pos.delete


def test_position_no_values_raises_error():
    with pytest.raises(ValueError) as exc:
        Position()

    assert (
        "One of cartesian, cartographicDegrees, cartographicRadians or reference must be given"
        in exc.exconly()
    )


def test_position_with_delete_has_nothing_else():
    expected_result = """{
    "delete": true
}"""
    pos_list = Position(delete=True, cartesian=[1, 2, 3])
    pos_val = Position(delete=True, cartesian=Cartesian3Value(values=[1, 2, 3]))

    assert str(pos_list) == str(pos_val) == expected_result


def test_position_has_given_epoch():
    expected_epoch = dt.datetime(2019, 6, 11, 12, 26, 58, tzinfo=dt.timezone.utc)

    pos = Position(epoch=expected_epoch, cartesian=[])

    assert pos.epoch == expected_epoch


def test_position_renders_epoch():
    expected_result = """{
    "epoch": "2019-03-20T12:00:00.000000Z",
    "cartesian": []
}"""
    pos = Position(
        epoch=dt.datetime(2019, 3, 20, 12, tzinfo=dt.timezone.utc), cartesian=[]
    )

    assert str(pos) == expected_result


def test_position_cartographic_degrees():
    expected_result = """{
    "cartographicDegrees": [
        10.0,
        20.0,
        0.0
    ]
}"""
    pos = Position(cartographicDegrees=[10.0, 20.0, 0.0])

    assert str(pos) == expected_result


def test_position_reference():
    expected_result = """{
    "reference": "satellite"
}"""
    pos = Position(reference="satellite")

    assert str(pos) == expected_result


def test_viewfrom_reference():
    expected_result = """{
    "reference": "satellite"
}"""
    v = ViewFrom(reference="satellite")

    assert str(v) == expected_result


def test_viewfrom_cartesian():
    expected_result = """{
    "cartesian": [
        -1000,
        0,
        300
    ]
}"""
    v = ViewFrom(cartesian=Cartesian3Value(values=[-1000, 0, 300]))

    assert str(v) == expected_result


def test_viewfrom_has_delete():
    v = ViewFrom(delete=True, cartesian=[])

    assert v.delete


def test_viewfrom_no_values_raises_error():
    with pytest.raises(ValueError) as exc:
        ViewFrom()

    assert "One of cartesian or reference must be given" in exc.exconly()


def test_single_interval_value():
    expected_result = """{
    "interval": "2019-01-01T00:00:00.000000Z/2019-01-02T00:00:00.000000Z",
    "boolean": true
}"""

    start = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    end = dt.datetime(2019, 1, 2, tzinfo=dt.timezone.utc)

    prop = IntervalValue(start=start, end=end, value=True)

    assert str(prop) == expected_result


def test_multiple_interval_value():
    expected_result = """[
    {
        "interval": "2019-01-01T00:00:00.000000Z/2019-01-02T00:00:00.000000Z",
        "boolean": true
    },
    {
        "interval": "2019-01-02T00:00:00.000000Z/2019-01-03T00:00:00.000000Z",
        "boolean": false
    }
]"""

    start0 = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    end0 = start1 = dt.datetime(2019, 1, 2, tzinfo=dt.timezone.utc)
    end1 = dt.datetime(2019, 1, 3, tzinfo=dt.timezone.utc)

    prop = Sequence(
        [
            IntervalValue(start=start0, end=end0, value=True),
            IntervalValue(start=start1, end=end1, value=False),
        ]
    )

    assert str(prop) == expected_result


def test_multiple_interval_decimal_value():
    expected_result = """[
    {
        "interval": "2019-01-01T01:02:03.456789Z/2019-01-02T01:02:03.456789Z",
        "boolean": true
    },
    {
        "interval": "2019-01-02T01:02:03.456789Z/2019-01-03T01:02:03.456789Z",
        "boolean": false
    }
]"""

    start0 = dt.datetime(2019, 1, 1, 1, 2, 3, 456789, tzinfo=dt.timezone.utc)
    end0 = start1 = dt.datetime(2019, 1, 2, 1, 2, 3, 456789, tzinfo=dt.timezone.utc)
    end1 = dt.datetime(2019, 1, 3, 1, 2, 3, 456789, tzinfo=dt.timezone.utc)

    prop = Sequence(
        [
            IntervalValue(start=start0, end=end0, value=True),
            IntervalValue(start=start1, end=end1, value=False),
        ]
    )

    assert str(prop) == expected_result


def test_orientation():
    expected_result = """{
    "unitQuaternion": [
        0,
        0,
        0,
        1
    ]
}"""

    result = Orientation(unitQuaternion=UnitQuaternionValue(values=[0, 0, 0, 1]))

    assert str(result) == expected_result


def test_model():
    expected_result = """{
    "gltf": "https://sandcastle.cesium.com/SampleData/models/CesiumAir/Cesium_Air.glb"
}"""

    result = Model(
        gltf="https://sandcastle.cesium.com/SampleData/models/CesiumAir/Cesium_Air.glb"
    )

    assert str(result) == expected_result


def test_bad_uri_raises_error():
    with pytest.raises(ValueError) as excinfo:
        Uri(uri="a")

    assert "uri must be a URL or a data URI" in excinfo.exconly()


def test_ellipsoid():
    expected_result = """{
    "radii": {
        "cartesian": [
            20.0,
            30.0,
            40.0
        ]
    },
    "fill": false,
    "outline": true
}"""

    ell = Ellipsoid(
        radii=EllipsoidRadii(cartesian=[20.0, 30.0, 40.0]), fill=False, outline=True
    )
    assert str(ell) == expected_result


def test_ellipsoid_parameters():
    expected_result = """{
    "radii": {
        "cartesian": [
            500000.0,
            500000.0,
            500000.0
        ]
    },
    "innerRadii": {
        "cartesian": [
            10000.0,
            10000.0,
            10000.0
        ]
    },
    "minimumClock": -15.0,
    "maximumClock": 15.0,
    "minimumCone": 75.0,
    "maximumCone": 105.0,
    "material": {
        "solidColor": {
            "rgba": [
                255,
                0,
                0,
                100
            ]
        }
    },
    "outline": true,
    "outlineColor": {
        "rgbaf": [
            0,
            0,
            0,
            1
        ]
    }
}"""

    ell = Ellipsoid(
        radii=EllipsoidRadii(cartesian=[500000.0, 500000.0, 500000.0]),
        innerRadii=EllipsoidRadii(cartesian=[10000.0, 10000.0, 10000.0]),
        minimumClock=-15.0,
        maximumClock=15.0,
        minimumCone=75.0,
        maximumCone=105.0,
        material=Material(
            solidColor=Color(rgba=[255, 0, 0, 100]),
        ),
        outline=True,
        outlineColor=Color(rgbaf=[0, 0, 0, 1]),
    )
    assert str(ell) == expected_result


def test_color_rgbaf_from_tuple():
    expected_result = """{
    "rgbaf": [
        0.127568,
        0.566949,
        0.550556,
        1.0
    ]
}"""
    tc = Color.from_tuple((0.127568, 0.566949, 0.550556, 1.0))
    assert str(tc) == expected_result


def test_color_rgba_from_tuple():
    expected_result = """{
    "rgba": [
        100,
        200,
        255,
        255
    ]
}"""
    tc = Color.from_tuple((100, 200, 255))
    assert str(tc) == expected_result



# noinspection PyPep8Naming
from __future__ import annotations

import datetime as dt

import attr

from .enums import InterpolationAlgorithms, ExtrapolationTypes


@attr.s(str=False, frozen=True, kw_only=True)
class Deletable:
    """A property whose value may be deleted."""

    delete: bool | None = attr.ib(default=None)


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Interpolatable:
    """A property whose value may be determined by interpolating.

    The interpolation happens over provided time-tagged samples.
    """

    epoch: dt.datetime | None = attr.ib(default=None)
    interpolationAlgorithm: InterpolationAlgorithms | None = attr.ib(default=None)
    interpolationDegree: int | None = attr.ib(default=None)
    forwardExtrapolationType: ExtrapolationTypes | None = attr.ib(default=None)
    backwardExtrapolationType: ExtrapolationTypes | None = attr.ib(default=None)


import pytest

from czmlpy.properties import Color
from czmlpy.types import RgbafValue, RgbaValue
from czmlpy.utils import get_color, get_color_list


def test_get_color_list_of_colors_rgba():
    expected_color = Color(
        rgba=RgbaValue(
            values=[
                "0000-00-00T00:00:00.000000Z",
                255,
                204,
                0,
                255,
                "9999-12-31T24:00:00.000000Z",
                255,
                204,
                0,
                255,
            ]
        )
    )
    assert (
        get_color_list(
            ["0000-00-00T00:00:00.000000Z", "9999-12-31T24:00:00.000000Z"],
            [[1.0, 0.8, 0.0, 1.0], 0xFFCC00FF],
        )
        == expected_color
    )
    assert (
        get_color_list(
            ["0000-00-00T00:00:00.000000Z", "9999-12-31T24:00:00.000000Z"],
            ["#ffcc00ff", 0xFFCC00],
        )
        == expected_color
    )


def test_get_color_list_of_colors_rgbaf():
    expected_color = Color(
        rgbaf=RgbafValue(
            values=[
                "0000-00-00T00:00:00.000000Z",
                1.0,
                0.8,
                0.0,
                1.0,
                "9999-12-31T24:00:00.000000Z",
                1.0,
                0.8,
                0.0,
                1.0,
            ]
        )
    )
    assert (
        get_color_list(
            ["0000-00-00T00:00:00.000000Z", "9999-12-31T24:00:00.000000Z"],
            [[1.0, 0.8, 0.0, 1.0], 0xFFCC00],
            rgbaf=True,
        )
        == expected_color
    )
    assert (
        get_color_list(
            ["0000-00-00T00:00:00.000000Z", "9999-12-31T24:00:00.000000Z"],
            [[255, 204, 0], 0xFFCC00FF],
            rgbaf=True,
        )
        == expected_color
    )


def test_get_color_list_of_colors_invalid():
    with pytest.raises(ValueError):
        get_color_list(
            ["0000-00-00T00:00:00.000000Z", "9999-12-31T24:00:00.000000Z"],
            [[300, 204, 0], -0xFFCC00FF],
            rgbaf=True,
        )


def test_get_color_rgba():
    expected_color = Color(rgba=RgbaValue(values=[255, 204, 0, 255]))

    assert get_color("#ffcc00") == expected_color
    assert get_color(0xFFCC00) == expected_color
    assert get_color("#ffcc00ff") == expected_color
    assert get_color(0xFFCC00FF) == expected_color
    assert get_color([255, 204, 0]) == expected_color
    assert get_color([255, 204, 0, 255]) == expected_color


def test_get_color_rgbaf():
    expected_color = Color(rgbaf=RgbafValue(values=[1.0, 0.8, 0.0, 1.0]))

    # TODO: Simplify after https://github.com/poliastro/czml3/issues/36
    assert get_color([1.0, 0.8, 0.0]) == expected_color
    assert get_color([1.0, 0.8, 0.0, 1.0]) == expected_color


@pytest.mark.parametrize("input", ["a", [0, 0, 0, 0, -300], [0.3, 0.3, 0.1, 1.0, 1.0]])
def test_get_color_invalid_input_raises_error(input):
    with pytest.raises(ValueError):
        get_color(input)



import datetime as dt
import json
import warnings
from enum import Enum
from json import JSONEncoder

import attr

from .constants import ISO8601_FORMAT_Z

NON_DELETE_PROPERTIES = ["id", "delete"]


class CZMLEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, BaseCZMLObject):
            return o.to_json()

        elif isinstance(o, Enum):
            return o.name

        elif isinstance(o, dt.datetime):
            return o.astimezone(dt.timezone.utc).strftime(ISO8601_FORMAT_Z)

        return super().default(o)


@attr.s(str=False, frozen=True)
class BaseCZMLObject:
    def __str__(self):
        return self.dumps(indent=4)

    def dumps(self, *args, **kwargs):
        if "cls" in kwargs:
            warnings.warn("Ignoring specified cls", UserWarning)

        kwargs["cls"] = CZMLEncoder
        return json.dumps(self, *args, **kwargs)

    def dump(self, fp, *args, **kwargs):
        for chunk in CZMLEncoder(*args, **kwargs).iterencode(self):
            fp.write(chunk)

    def to_json(self):
        if getattr(self, "delete", False):
            properties_list = NON_DELETE_PROPERTIES
        else:
            properties_list = list(attr.asdict(self).keys())

        obj_dict = {}
        for property_name in properties_list:
            if getattr(self, property_name, None) is not None:
                obj_dict[property_name] = getattr(self, property_name)

        return obj_dict



import datetime as dt

import astropy.time
import pytest
from dateutil.tz import tzoffset

from czmlpy.types import (
    Cartesian3Value,
    CartographicDegreesListValue,
    CartographicRadiansListValue,
    DistanceDisplayConditionValue,
    FontValue,
    NearFarScalarValue,
    ReferenceValue,
    RgbafValue,
    RgbaValue,
    TimeInterval,
    UnitQuaternionValue,
    format_datetime_like,
)


def test_invalid_near_far_scalar_value():
    with pytest.raises(ValueError) as excinfo:
        NearFarScalarValue(values=[0, 3.2, 1, 4, 2, 1])

    assert "Input values must have either 4 or N * 5 values, " in excinfo.exconly()


def test_distance_display_condition():
    expected_result = """[
    0,
    150,
    15000000,
    300,
    10000,
    15000000,
    600,
    150,
    15000000
]"""
    dist = DistanceDisplayConditionValue(
        values=[0, 150, 15000000, 300, 10000, 15000000, 600, 150, 15000000]
    )
    assert str(dist) == expected_result


def test_cartographic_radian_list():
    expected_result = """[
    0,
    1,
    0
]"""
    car = CartographicRadiansListValue(values=[0, 1, 0])
    assert str(car) == expected_result


def test_invalid_cartograpic_radian_list():

    with pytest.raises(ValueError) as excinfo:
        CartographicRadiansListValue(values=[1])
    assert (
        "Invalid values. Input values should be arrays of size 3 * N"
        in excinfo.exconly()
    )


def test_cartograpic_degree_list():
    expected_result = """[
    15,
    25,
    50
]"""
    car = CartographicDegreesListValue(values=[15, 25, 50])
    assert str(car) == expected_result


def test_invalid_cartograpic_degree_list():

    with pytest.raises(ValueError) as excinfo:
        CartographicDegreesListValue(values=[15, 25, 50, 30])
    assert (
        "Invalid values. Input values should be arrays of size 3 * N"
        in excinfo.exconly()
    )


@pytest.mark.parametrize("values", [[2, 2], [5, 5, 5, 5, 5]])
def test_bad_cartesian_raises_error(values):
    with pytest.raises(ValueError) as excinfo:
        Cartesian3Value(values=values)

    assert "Input values must have either 3 or N * 4 values" in excinfo.exconly()


def test_reference_value():
    expected_result = '"id#property"'
    reference = ReferenceValue(string="id#property")

    assert str(reference) == expected_result


def test_invalid_reference_value():
    with pytest.raises(ValueError) as excinfo:
        ReferenceValue(string="id")

    assert (
        "Invalid reference string format. Input must be of the form id#property"
        in excinfo.exconly()
    )


def test_font_value():
    expected_result = '"20px sans-serif"'
    font = FontValue(font="20px sans-serif")

    assert str(font) == expected_result


def test_font_property_value():
    expected_result = "20px sans-serif"
    font = FontValue(font="20px sans-serif")

    assert font.font == expected_result


def test_bad_rgba_size_values_raises_error():
    with pytest.raises(ValueError) as excinfo:
        RgbaValue(values=[0, 0, 255])

    assert "Input values must have either 4 or N * 5 values, " in excinfo.exconly()


def test_bad_rgba_4_values_raises_error():
    with pytest.raises(ValueError) as excinfo:
        RgbaValue(values=[256, 0, 0, 255])

    assert "Color values must be integers in the range 0-255." in excinfo.exconly()


def test_bad_rgba_5_color_values_raises_error():

    with pytest.raises(ValueError) as excinfo:
        RgbaValue(values=[0, 0.1, 0.3, 0.3, 255])

    assert "Color values must be integers in the range 0-255." in excinfo.exconly()


def test_bad_rgbaf_size_values_raises_error():
    with pytest.raises(ValueError) as excinfo:
        RgbafValue(values=[0, 0, 0.1])

    assert "Input values must have either 4 or N * 5 values, " in excinfo.exconly()


def test_bad_rgbaf_4_values_raises_error():
    with pytest.raises(ValueError) as excinfo:
        RgbafValue(values=[0.3, 0, 0, 1.4])

    assert "Color values must be floats in the range 0-1." in excinfo.exconly()


def test_bad_rgbaf_5_color_values_raises_error():

    with pytest.raises(ValueError) as excinfo:
        RgbafValue(values=[0, 0.1, 0.3, 0.3, 255])

    assert "Color values must be floats in the range 0-1." in excinfo.exconly()


def test_default_time_interval():
    expected_result = '"0000-00-00T00:00:00Z/9999-12-31T24:00:00Z"'
    time_interval = TimeInterval()

    assert str(time_interval) == expected_result


def test_custom_time_interval():
    tz = tzoffset("UTC+02", dt.timedelta(hours=2))
    start = dt.datetime(2019, 1, 1, 12, 0, tzinfo=dt.timezone.utc)
    end = dt.datetime(2019, 9, 2, 23, 59, 59, tzinfo=tz)

    expected_result = '"2019-01-01T12:00:00.000000Z/2019-09-02T21:59:59.000000Z"'

    time_interval = TimeInterval(start=start, end=end)

    assert str(time_interval) == expected_result


def test_bad_time_raises_error():
    with pytest.raises(ValueError):
        format_datetime_like("2019/01/01")


@pytest.mark.xfail
def test_astropy_time_retains_input_format():
    # It would be nice to recover the input format,
    # but it's difficult without conditionally depending on Astropy
    expected_result = "2012-03-15T10:16:06.97400000000198Z"
    time = astropy.time.Time(expected_result)

    result = format_datetime_like(time)

    assert result == expected_result


def test_astropy_time_format():
    expected_result = "2012-03-15T10:16:06.974Z"
    time = astropy.time.Time("2012-03-15T10:16:06.97400000000198Z")

    result = format_datetime_like(time)

    assert result == expected_result


def test_quaternion_value():
    expected_result = """[
    0,
    0,
    0,
    1
]"""

    result = UnitQuaternionValue(values=[0, 0, 0, 1])

    assert str(result) == expected_result



from __future__ import annotations

import attr
from w3lib.url import is_url, parse_data_uri

from .base import BaseCZMLObject
from .common import Deletable, Interpolatable
from .enums import (
    ClockRanges,
    ClockSteps,
    HorizontalOrigins,
    LabelStyles,
    StripeOrientations,
    VerticalOrigins,
)
from .types import RgbafValue, RgbaValue


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class HasAlignment:
    """A property that can be horizontally or vertically aligned."""

    horizontalOrigin: HorizontalOrigins | None = attr.ib(default=None)
    verticalOrigin: VerticalOrigins | None = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Material(BaseCZMLObject):
    """A definition of how a surface is colored or shaded."""

    solidColor = attr.ib(default=None)
    image = attr.ib(default=None)
    grid = attr.ib(default=None)
    stripe = attr.ib(default=None)
    checkerboard = attr.ib(default=None)
    polylineOutline = attr.ib(default=None)  # NOTE: Not present in documentation


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineOutline(BaseCZMLObject):
    """A definition of how a surface is colored or shaded."""

    color = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineOutlineMaterial(BaseCZMLObject):
    """A definition of the material wrapper for a polyline outline."""

    polylineOutline = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineGlow(BaseCZMLObject):
    """A definition of how a glowing polyline appears."""

    color = attr.ib(default=None)
    glowPower = attr.ib(default=None)
    taperPower = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineGlowMaterial(BaseCZMLObject):
    """A material that fills the surface of a line with a glowing color."""

    polylineGlow = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineArrow(BaseCZMLObject):
    """A definition of how a polyline arrow appears."""

    color = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineArrowMaterial(BaseCZMLObject):
    """A material that fills the surface of a line with an arrow."""

    polylineArrow = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineDash(BaseCZMLObject):
    """A definition of how a polyline should be dashed with two colors."""

    color = attr.ib(default=None)
    gapColor = attr.ib(default=None)
    dashLength = attr.ib(default=None)
    dashPattern = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineDashMaterial(BaseCZMLObject):
    """A material that provides a how a polyline should be dashed."""

    polylineDash = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PolylineMaterial(BaseCZMLObject):
    """A definition of how a surface is colored or shaded."""

    solidColor = attr.ib(default=None)
    image = attr.ib(default=None)
    grid = attr.ib(default=None)
    stripe = attr.ib(default=None)
    checkerboard = attr.ib(default=None)
    polylineDash = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class SolidColorMaterial(BaseCZMLObject):
    """A material that fills the surface with a solid color."""

    color = attr.ib(default=None)

    @classmethod
    def from_list(cls, color):
        return cls(color=Color.from_list(color))


@attr.s(str=False, frozen=True, kw_only=True)
class GridMaterial(BaseCZMLObject):
    """A material that fills the surface with a two-dimensional grid."""

    color = attr.ib(default=None)
    cellAlpha = attr.ib(default=0.1)
    lineCount = attr.ib(default=[8, 8])
    lineThickness = attr.ib(default=[1.0, 1.0])
    lineOffset = attr.ib(default=[0.0, 0.0])


@attr.s(str=False, frozen=True, kw_only=True)
class StripeMaterial(BaseCZMLObject):
    """A material that fills the surface with alternating colors."""

    orientation = attr.ib(default=StripeOrientations.HORIZONTAL)
    evenColor = attr.ib(default=None)
    oddColor = attr.ib(default=None)
    offset = attr.ib(default=0.0)
    repeat = attr.ib(default=1.0)


@attr.s(str=False, frozen=True, kw_only=True)
class CheckerboardMaterial(BaseCZMLObject):
    """A material that fills the surface with alternating colors."""

    evenColor = attr.ib(default=None)
    oddColor = attr.ib(default=None)
    repeat = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class ImageMaterial(BaseCZMLObject):
    """A material that fills the surface with an image."""

    image = attr.ib(default=None)
    repeat = attr.ib(default=[1, 1])
    color = attr.ib(default=None)
    transparent = attr.ib(default=False)


@attr.s(str=False, frozen=True, kw_only=True)
class Color(BaseCZMLObject, Interpolatable, Deletable):
    """A color. The color can optionally vary over time."""

    rgba = attr.ib(default=None)
    rgbaf = attr.ib(default=None)

    @classmethod
    def is_valid(cls, color):
        """Determines if the input is a valid color"""
        # [R, G, B] or [R, G, B, A]
        if (
            isinstance(color, (list, tuple))
            and all([issubclass(type(v), int) for v in color])
            and (3 <= len(color) <= 4)
        ):
            return all(0 <= v <= 255 for v in color)
        # [r, g, b] or [r, g, b, a] (float)
        elif (
            isinstance(color, (list, tuple))
            and all([issubclass(type(v), float) for v in color])
            and (3 <= len(color) <= 4)
        ):
            return all(0 <= v <= 1 for v in color)
        # Hexadecimal RGBA
        elif issubclass(type(color), int):
            return 0 <= color <= 0xFFFFFFFF
        # RGBA string
        elif isinstance(color, str):
            try:
                n = int(color.rsplit("#")[-1], 16)
                return 0 <= n <= 0xFFFFFFFF
            except ValueError:
                return False
        return False

    @classmethod
    def from_list(cls, color):
        if all(issubclass(type(v), int) for v in color):
            if len(color) == 3:
                color = color + [255]
            else:
                color = color[:]

            return cls(rgba=RgbaValue(values=color))
        else:
            if len(color) == 3:
                color = color + [1.0]
            else:
                color = color[:]

            return cls(rgbaf=RgbafValue(values=color))

    @classmethod
    def from_tuple(cls, color):
        return cls.from_list(list(color))

    @classmethod
    def from_hex(cls, color):
        if color > 0xFFFFFF:
            values = [
                (color & 0xFF000000) >> 24,
                (color & 0x00FF0000) >> 16,
                (color & 0x0000FF00) >> 8,
                (color & 0x000000FF) >> 0,
            ]
        else:
            values = [
                (color & 0xFF0000) >> 16,
                (color & 0x00FF00) >> 8,
                (color & 0x0000FF) >> 0,
                0xFF,
            ]

        return cls.from_list(values)

    @classmethod
    def from_str(cls, color):
        return cls.from_hex(int(color.rsplit("#")[-1], 16))


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Position(BaseCZMLObject, Interpolatable, Deletable):
    """Defines a position. The position can optionally vary over time."""

    referenceFrame = attr.ib(default=None)
    cartesian = attr.ib(default=None)
    cartographicRadians = attr.ib(default=None)
    cartographicDegrees = attr.ib(default=None)
    cartesianVelocity = attr.ib(default=None)
    reference = attr.ib(default=None)

    def __attrs_post_init__(self):
        if all(
            val is None
            for val in (
                self.cartesian,
                self.cartographicDegrees,
                self.cartographicRadians,
                self.cartesianVelocity,
                self.reference,
            )
        ):
            raise ValueError(
                "One of cartesian, cartographicDegrees, cartographicRadians or reference must be given"
            )


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class ViewFrom(BaseCZMLObject, Interpolatable, Deletable):
    """suggested initial camera position offset when tracking this object.

    ViewFrom can optionally vary over time."""

    cartesian = attr.ib(default=None)
    reference = attr.ib(default=None)

    def __attrs_post_init__(self):
        if all(val is None for val in (self.cartesian, self.reference)):
            raise ValueError("One of cartesian or reference must be given")


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Billboard(BaseCZMLObject, HasAlignment):
    """A billboard, or viewport-aligned image.

    The billboard is positioned in the scene by the position property.
    A billboard is sometimes called a marker.
    """

    image = attr.ib()
    show = attr.ib(default=None)
    scale = attr.ib(default=None)
    eyeOffset = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class EllipsoidRadii(BaseCZMLObject, Interpolatable, Deletable):
    """The radii of an ellipsoid."""

    cartesian = attr.ib(default=None)
    reference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Corridor(BaseCZMLObject):
    """A corridor , which is a shape defined by a centerline and width that conforms to the
    curvature of the body shape. It can can optionally be extruded into a volume."""

    positions = attr.ib()
    show = attr.ib(default=None)
    width = attr.ib()
    height = attr.ib(default=None)
    heightReference = attr.ib(default=None)
    extrudedHeight = attr.ib(default=None)
    extrudedHeightReference = attr.ib(default=None)
    cornerType = attr.ib(default=None)
    granularity = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)
    outline = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    shadows = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)
    classificationType = attr.ib(default=None)
    zIndex = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Cylinder(BaseCZMLObject):
    """A cylinder, which is a special cone defined by length, top and bottom radius."""

    length = attr.ib()
    show = attr.ib(default=None)
    topRadius = attr.ib()
    bottomRadius = attr.ib()
    heightReference = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)
    outline = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    numberOfVerticalLines = attr.ib(default=None)
    slices = attr.ib(default=None)
    shadows = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Ellipse(BaseCZMLObject):
    """An ellipse, which is a close curve, on or above Earth's surface."""

    semiMajorAxis = attr.ib()
    semiMinorAxis = attr.ib()
    show = attr.ib(default=None)
    height = attr.ib(default=None)
    heightReference = attr.ib(default=None)
    extrudedHeight = attr.ib(default=None)
    extrudedHeightReference = attr.ib(default=None)
    rotation = attr.ib(default=None)
    stRotation = attr.ib(default=None)
    granularity = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)
    outline = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    numberOfVerticalLines = attr.ib(default=None)
    shadows = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)
    classificationType = attr.ib(default=None)
    zIndex = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Polygon(BaseCZMLObject):
    """A polygon, which is a closed figure on the surface of the Earth."""

    positions = attr.ib()
    show = attr.ib(default=None)
    arcType = attr.ib(default=None)
    granularity = attr.ib(default=None)
    material = attr.ib(default=None)
    shadows = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)
    classificationType = attr.ib(default=None)
    zIndex = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Polyline(BaseCZMLObject):
    """A polyline, which is a line in the scene composed of multiple segments."""

    positions = attr.ib()
    show = attr.ib(default=None)
    arcType = attr.ib(default=None)
    width = attr.ib(default=None)
    granularity = attr.ib(default=None)
    material = attr.ib(default=None)
    followSurface = attr.ib(default=None)
    shadows = attr.ib(default=None)
    depthFailMaterial = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)
    clampToGround = attr.ib(default=None)
    classificationType = attr.ib(default=None)
    zIndex = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class ArcType(BaseCZMLObject, Deletable):
    """The type of an arc."""

    arcType = attr.ib(default=None)
    reference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class ShadowMode(BaseCZMLObject, Deletable):
    """Whether or not an object casts or receives shadows from each light source when shadows are enabled."""

    shadowMode = attr.ib(default=None)
    referenec = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class ClassificationType(BaseCZMLObject, Deletable):
    """Whether a classification affects terrain, 3D Tiles, or both."""

    classificationType = attr.ib(default=None)
    reference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class DistanceDisplayCondition(BaseCZMLObject, Interpolatable, Deletable):
    """Indicates the visibility of an object based on the distance to the camera."""

    distanceDisplayCondition = attr.ib(default=None)
    reference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class PositionList(BaseCZMLObject, Deletable):
    """A list of positions."""

    referenceFrame = attr.ib(default=None)
    cartesian = attr.ib(default=None)
    cartographicRadians = attr.ib(default=None)
    cartographicDegrees = attr.ib(default=None)
    references = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class DirectionList(BaseCZMLObject, Deletable):
    """A list of positions."""

    spherical = attr.ib(default=None)
    unitSpherical = attr.ib(default=None)
    cartesian = attr.ib(default=None)
    unitCartesian = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Ellipsoid(BaseCZMLObject):
    """A closed quadric surface that is a three-dimensional analogue of an ellipse."""

    radii = attr.ib()
    innerRadii = attr.ib(default=None)
    minimumClock = attr.ib(default=None)
    maximumClock = attr.ib(default=None)
    minimumCone = attr.ib(default=None)
    maximumCone = attr.ib(default=None)
    show = attr.ib(default=None)
    heightReference = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)
    outline = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    stackPartitions = attr.ib(default=None)
    slicePartitions = attr.ib(default=None)
    subdivisions = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Box(BaseCZMLObject):
    """A box, which is a closed rectangular cuboid."""

    show = attr.ib(default=None)
    dimensions = attr.ib(default=None)
    heightReference = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)
    outline = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    shadows = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class BoxDimensions(BaseCZMLObject, Interpolatable):
    """The width, depth, and height of a box."""

    cartesian = attr.ib(default=None)
    reference = attr.ib(default=None)


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Rectangle(BaseCZMLObject, Interpolatable, Deletable):
    """A cartographic rectangle, which conforms to the curvature of the globe and
    can be placed on the surface or at altitude and can optionally be extruded into a volume."""

    coordinates = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class RectangleCoordinates(BaseCZMLObject, Interpolatable, Deletable):
    """A set of coordinates describing a cartographic rectangle on the surface of the ellipsoid."""

    reference = attr.ib(default=None)
    wsen = attr.ib(default=None)
    wsenDegrees = attr.ib(default=None)

    def __attrs_post_init__(self):
        if all(val is None for val in (self.wsen, self.wsenDegrees)):
            raise ValueError(
                "One of cartesian, cartographicDegrees or cartographicRadians must be given"
            )


@attr.s(str=False, frozen=True, kw_only=True)
class EyeOffset(BaseCZMLObject, Deletable):
    """An offset in eye coordinates which can optionally vary over time.

    Eye coordinates are a left-handed coordinate system
    where the X-axis points toward the viewer's right,
    the Y-axis poitns up, and the Z-axis points into the screen.

    """

    cartesian = attr.ib(default=None)
    reference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class HeightReference(BaseCZMLObject, Deletable):
    """The height reference of an object, which indicates if the object's position is relative to terrain or not."""

    heightReference = attr.ib(default=None)
    reference = attr.ib(default=None)


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Clock(BaseCZMLObject):
    """Initial settings for a simulated clock when a document is loaded.

    The start and stop time are configured using the interval property.

    """

    currentTime = attr.ib(default=None)
    multiplier = attr.ib(default=1.0)
    range = attr.ib(default=ClockRanges.LOOP_STOP)
    step = attr.ib(default=ClockSteps.SYSTEM_CLOCK_MULTIPLIER)


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Path(BaseCZMLObject):
    """A path, which is a polyline defined by the motion of an object over time.

    The possible vertices of the path are specified by the position property.
    Note that because clients cannot render a truly infinite path,
    the path must be limited,
    either by defining availability for this object,
    or by using the leadTime and trailTime properties.

    """

    show = attr.ib(default=None)
    leadTime = attr.ib(default=None)
    trailTime = attr.ib(default=None)
    width = attr.ib(default=1.0)
    resolution = attr.ib(default=60.0)
    material = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Point(BaseCZMLObject):
    """A point, or viewport-aligned circle."""

    show = attr.ib(default=None)
    pixelSize = attr.ib(default=None)
    heightReference = attr.ib(default=None)
    color = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    scaleByDistance = attr.ib(default=None)
    translucencyByDistance = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)
    disableDepthTestDistance = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class TileSet(BaseCZMLObject):
    """A 3D Tiles tileset."""

    show = attr.ib(default=None)
    uri = attr.ib()
    maximumScreenSpaceError = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Wall(BaseCZMLObject):
    """A two-dimensional wall defined as a line strip and optional maximum and minimum heights.
    It conforms to the curvature of the globe and can be placed along the surface or at altitude."""

    show = attr.ib(default=None)
    positions = attr.ib()
    minimumHeights = attr.ib(default=None)
    maximumHeights = attr.ib(default=None)
    granularity = attr.ib(default=None)
    fill = attr.ib(default=None)
    material = attr.ib(default=None)
    outline = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=None)
    shadows = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class NearFarScalar(BaseCZMLObject, Interpolatable, Deletable):
    """A numeric value which will be linearly interpolated between two values based on an object's distance from the
     camera, in eye coordinates.

    The computed value will interpolate between the near value and the far value while the camera distance falls
    between the near distance and the far distance, and will be clamped to the near or far value while the distance is
    less than the near distance or greater than the far distance, respectively.
    """

    nearFarScalar = attr.ib(default=None)
    reference = attr.ib(default=None)


# noinspection PyPep8Naming
@attr.s(str=False, frozen=True, kw_only=True)
class Label(BaseCZMLObject, HasAlignment):
    """A string of text."""

    show = attr.ib(default=True)
    text = attr.ib(default=None)
    font = attr.ib(default=None)
    style = attr.ib(default=LabelStyles.FILL)
    scale = attr.ib(default=None)
    showBackground = attr.ib(default=None)
    backgroundColor = attr.ib(default=None)
    fillColor = attr.ib(default=None)
    outlineColor = attr.ib(default=None)
    outlineWidth = attr.ib(default=1.0)
    pixelOffset = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Orientation(BaseCZMLObject, Interpolatable, Deletable):
    """Defines an orientation.

    An orientation is a rotation that takes a vector expressed in the "body" axes of the object
    and transforms it to the Earth fixed axes.

    """

    unitQuaternion = attr.ib(default=None)
    reference = attr.ib(default=None)
    velocityReference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Model(BaseCZMLObject):
    """A 3D model."""

    show = attr.ib(default=None)
    gltf = attr.ib()
    scale = attr.ib(default=None)
    minimumPixelSize = attr.ib(default=None)
    maximumScale = attr.ib(default=None)
    incrementallyLoadTextures = attr.ib(default=None)
    runAnimations = attr.ib(default=None)
    shadows = attr.ib(default=None)
    heightReference = attr.ib(default=None)
    silhouetteColor = attr.ib(default=None)
    silhouetteSize = attr.ib(default=None)
    color = attr.ib(default=None)
    colorBlendMode = attr.ib(default=None)
    colorBlendAmount = attr.ib(default=None)
    distanceDisplayCondition = attr.ib(default=None)
    nodeTransformations = attr.ib(default=None)
    articulations = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Uri(BaseCZMLObject, Deletable):
    """A URI value.

    The URI can optionally vary with time.
    """

    uri = attr.ib(default=None)

    @uri.validator
    def _check_uri(self, attribute, value):
        try:
            parse_data_uri(value)
        except ValueError as e:
            if not is_url(value):
                raise ValueError("uri must be a URL or a data URI") from e

    def to_json(self):
        return self.uri


@attr.s(str=False, frozen=True, kw_only=True)
class Double(BaseCZMLObject, Interpolatable, Deletable):

    number = attr.ib(default=None)
    reference = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Integer(BaseCZMLObject, Interpolatable, Deletable):

    number = attr.ib(default=None)
    reference = attr.ib(default=None)



@attr.s(str=False, frozen=True, kw_only=True)
class ConicSensor(BaseCZMLObject):

    show = attr.ib(default=None)
    innerHalfAngle = attr.ib(default=None)
    outerHalfAngle = attr.ib(default=None)
    minimumClockAngle = attr.ib(default=None)
    maximumClockAngle = attr.ib(default=None)
    radius = attr.ib(default=None)
    showIntersection = attr.ib(default=None)
    intersectionColor = attr.ib(default=None)
    intersectionWidth = attr.ib(default=None)
    showLateralSurfaces = attr.ib(default=None)
    lateralSurfaceMaterial = attr.ib(default=None)
    showEllipsoidSurfaces = attr.ib(default=None)
    ellipsoidSurfaceMaterial = attr.ib(default=None)
    showEllipsoidHorizonSurfaces = attr.ib(default=None)
    ellipsoidHorizonSurfaceMaterial = attr.ib(default=None)
    showDomeSurfaces = attr.ib(default=None)
    domeSurfaceMaterial = attr.ib(default=None)
    portionToDisplay = attr.ib(default=None)
    environmentConstraint = attr.ib(default=None)
    showEnvironmentOcclusion = attr.ib(default=None)
    environmentOcclusionMaterial = attr.ib(default=None)
    showEnvironmentIntersection = attr.ib(default=None)
    environmentIntersectionColor = attr.ib(default=None)
    environmentIntersectionWidth = attr.ib(default=None)
    showThroughEllipsoid = attr.ib(default=None)
    showViewshed = attr.ib(default=None)
    viewshedVisibleColor = attr.ib(default=None)
    viewshedOccludedColor = attr.ib(default=None)
    viewshedResolution = attr.ib(default=None)

@attr.s(str=False, frozen=True, kw_only=True)
class CustomPatternSensor(BaseCZMLObject):

    show = attr.ib(default=None)
    directions = attr.ib(default=None)
    radius = attr.ib(default=None)
    showIntersection = attr.ib(default=None)
    intersectionColor = attr.ib(default=None)
    intersectionWidth = attr.ib(default=None)
    showLateralSurfaces = attr.ib(default=None)
    lateralSurfaceMaterial = attr.ib(default=None)
    showEllipsoidSurfaces = attr.ib(default=None)
    ellipsoidSurfaceMaterial = attr.ib(default=None)
    showEllipsoidHorizonSurfaces = attr.ib(default=None)
    ellipsoidHorizonSurfaceMaterial = attr.ib(default=None)
    showDomeSurfaces = attr.ib(default=None)
    domeSurfaceMaterial = attr.ib(default=None)
    portionToDisplay = attr.ib(default=None)
    environmentConstraint = attr.ib(default=None)
    showEnvironmentOcclusion = attr.ib(default=None)
    environmentOcclusionMaterial = attr.ib(default=None)
    showEnvironmentIntersection = attr.ib(default=None)
    environmentIntersectionColor = attr.ib(default=None)
    environmentIntersectionWidth = attr.ib(default=None)
    showThroughEllipsoid = attr.ib(default=None)
    showViewshed = attr.ib(default=None)
    viewshedVisibleColor = attr.ib(default=None)
    viewshedOccludedColor = attr.ib(default=None)
    viewshedResolution = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class RectangularSensor(BaseCZMLObject):

    show = attr.ib(default=None)
    xHalfAngle = attr.ib(default=None)
    yHalfAngle = attr.ib(default=None)
    radius = attr.ib(default=None)
    showIntersection = attr.ib(default=None)
    intersectionColor = attr.ib(default=None)
    intersectionWidth = attr.ib(default=None)
    showLateralSurfaces = attr.ib(default=None)
    lateralSurfaceMaterial = attr.ib(default=None)
    showEllipsoidSurfaces = attr.ib(default=None)
    ellipsoidSurfaceMaterial = attr.ib(default=None)
    showEllipsoidHorizonSurfaces = attr.ib(default=None)
    ellipsoidHorizonSurfaceMaterial = attr.ib(default=None)
    showDomeSurfaces = attr.ib(default=None)
    domeSurfaceMaterial = attr.ib(default=None)
    portionToDisplay = attr.ib(default=None)
    environmentConstraint = attr.ib(default=None)
    showEnvironmentOcclusion = attr.ib(default=None)
    environmentOcclusionMaterial = attr.ib(default=None)
    showEnvironmentIntersection = attr.ib(default=None)
    environmentIntersectionColor = attr.ib(default=None)
    environmentIntersectionWidth = attr.ib(default=None)
    showThroughEllipsoid = attr.ib(default=None)
    showViewshed = attr.ib(default=None)
    viewshedVisibleColor = attr.ib(default=None)
    viewshedOccludedColor = attr.ib(default=None)
    viewshedResolution = attr.ib(default=None)



from uuid import uuid4

import attr

from .base import BaseCZMLObject
from .types import Sequence

CZML_VERSION = "1.0"


@attr.s(str=False, frozen=True, kw_only=True)
class Preamble(BaseCZMLObject):
    """The preamble packet."""

    id = attr.ib(init=False, default="document")

    version = attr.ib(default=CZML_VERSION)
    name = attr.ib(default=None)
    description = attr.ib(default=None)
    clock = attr.ib(default=None)


@attr.s(str=False, frozen=True, kw_only=True)
class Packet(BaseCZMLObject):
    """A CZML Packet.

    See https://github.com/AnalyticalGraphicsInc/czml-writer/wiki/Packet
    for further information.
    """

    id = attr.ib(factory=lambda: str(uuid4()))
    delete = attr.ib(default=None)
    name = attr.ib(default=None)
    parent = attr.ib(default=None)
    description = attr.ib(default=None)
    availability = attr.ib(default=None)
    properties = attr.ib(default=None)
    position = attr.ib(default=None)
    orientation = attr.ib(default=None)
    viewFrom = attr.ib(default=None)
    billboard = attr.ib(default=None)
    box = attr.ib(default=None)
    corridor = attr.ib(default=None)
    cylinder = attr.ib(default=None)
    ellipse = attr.ib(default=None)
    ellipsoid = attr.ib(default=None)
    label = attr.ib(default=None)
    model = attr.ib(default=None)
    path = attr.ib(default=None)
    point = attr.ib(default=None)
    polygon = attr.ib(default=None)
    polyline = attr.ib(default=None)
    rectangle = attr.ib(default=None)
    tileset = attr.ib(default=None)
    wall = attr.ib(default=None)
    agi_conicSensor = attr.ib(default=None)
    agi_customPatternSensor = attr.ib(default=None)
    agi_rectangularSensor = attr.ib(default=None)
    agi_fan = attr.ib(default=None)
    agi_vector  = attr.ib(default=None)


@attr.s(str=False, frozen=True)
class Document(Sequence):
    """A CZML document, consisting on a list of packets."""

    @property
    def packets(self):
        return self._values



from enum import Enum, auto


class InterpolationAlgorithms(Enum):
    """The interpolation algorithm to use when interpolating."""

    LINEAR = auto()
    LAGRANGE = auto()
    HERMITE = auto()


class ExtrapolationTypes(Enum):
    """The type of extrapolation to perform when a value is requested at a time after any available samples."""

    NONE = auto()
    HOLD = auto()
    EXTRAPOLATE = auto()


class ReferenceFrames(Enum):
    """The reference frame in which cartesian positions are specified."""

    FIXED = auto()
    INERTIAL = auto()


class LabelStyles(Enum):
    """The style of a label."""

    FILL = auto()
    OUTLINE = auto()
    FILL_AND_OUTLINE = auto()


class ClockRanges(Enum):
    """The behavior of a clock when its current time reaches its start or end time."""

    UNBOUNDED = auto()
    CLAMPED = auto()
    LOOP_STOP = auto()


class ClockSteps(Enum):
    TICK_DEPENDENT = auto()
    SYSTEM_CLOCK_MULTIPLIER = auto()
    SYSTEM_CLOCK = auto()


class VerticalOrigins(Enum):
    BASELINE = auto()
    BOTTOM = auto()
    CENTER = auto()
    TOP = auto()


class HorizontalOrigins(Enum):
    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()


class HeightReferences(Enum):
    NONE = auto()
    CLAMP_TO_GROUND = auto()
    RELATIVE_TO_GROUND = auto()


class ColorBlendModes(Enum):
    HIGHLIGHT = auto()
    REPLACE = auto()
    MIX = auto()


class ShadowModes(Enum):
    DISABLED = auto()
    ENABLED = auto()
    CAST_ONLY = auto()
    RECEIVE_ONLY = auto()


class ClassificationTypes(Enum):
    TERRAIN = auto()
    CESIUM_3D_TILE = auto()
    BOTH = auto()


class ArcTypes(Enum):
    NONE = auto()
    GEODESIC = auto()
    RHUMB = auto()


class StripeOrientations(Enum):
    HORIZONTAL = auto()
    VERTICAL = auto()



from czmlpy.core import Preamble


def test_empty_preamble_has_id_and_version():
    expected_result = """{
    "id": "document",
    "version": "1.0"
}"""

    packet = Preamble()

    assert str(packet) == expected_result



import base64
import os
import tempfile

import pytest

from czmlpy import Document, Packet, Preamble
from czmlpy.properties import ImageMaterial, Material, Rectangle, RectangleCoordinates


@pytest.fixture
def image():
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), "smiley.png")
    with open(filename, "rb") as fp:
        data = fp.read()

    base64_data = base64.b64encode(data)
    return base64_data.decode("utf-8")


def test_rectangle_coordinates_invalid_if_nothing_given():
    with pytest.raises(ValueError) as excinfo:
        RectangleCoordinates()

    assert (
        "One of cartesian, cartographicDegrees or cartographicRadians must be given"
        in excinfo.exconly()
    )


def test_packet_rectangles(image):
    wsen = [20, 40, 21, 41]

    expected_result = """{{
    "id": "id_00",
    "rectangle": {{
        "coordinates": {{
            "wsenDegrees": [
                {},
                {},
                {},
                {}
            ]
        }},
        "fill": true,
        "material": {{
            "image": {{
                "image": "data:image/png;base64,{}",
                "transparent": true
            }}
        }}
    }}
}}""".format(
        *wsen, image
    )

    rectangle_packet = Packet(
        id="id_00",
        rectangle=Rectangle(
            coordinates=RectangleCoordinates(wsenDegrees=wsen),
            fill=True,
            material=Material(
                image=ImageMaterial(
                    transparent=True,
                    repeat=None,
                    image="data:image/png;base64," + image,
                ),
            ),
        ),
    )

    assert str(rectangle_packet) == expected_result


def test_make_czml_png_rectangle_file(image):
    wsen = [20, 40, 21, 41]

    rectangle_packet = Packet(
        id="id_00",
        rectangle=Rectangle(
            coordinates=RectangleCoordinates(wsenDegrees=wsen),
            fill=True,
            material=Material(
                image=ImageMaterial(
                    transparent=True,
                    repeat=None,
                    image="data:image/png;base64," + image,
                ),
            ),
        ),
    )

    with tempfile.NamedTemporaryFile(mode="w", suffix=".czml") as out_file:
        out_file.write(str(Document([Preamble(), rectangle_packet])))
        exists = os.path.isfile(out_file.name)

        # TODO: Should we be testing something else?
        assert exists



from uuid import uuid4

import attr

from .core import Document, Preamble

TERRAIN = {
    "Cesium": "Cesium.createWorldTerrain()",
    "Ellipsoid": "new Cesium.EllipsoidTerrainProvider()",
}

IMAGERY = {
    "Bing_Aerial": "Cesium.createWorldImagery()",
    "OSM": "new Cesium.OpenStreetMapImageryProvider()",
}

CESIUM_TPL = """
<link rel="stylesheet" href="https://cesium.com/downloads/cesiumjs/releases/{cesium_version}/Build/Cesium/Widgets/widgets.css" type="text/css">
<div id="cesiumContainer-{container_id}" style="width:100%; height:{widget_height};"></div>
<script type="text/javascript">
{script}
</script>"""

SCRIPT_TPL = """
require.config({{
    paths: {{
        'cesium': 'https://cesium.com/downloads/cesiumjs/releases/{cesium_version}/Build/Cesium/Cesium'
    }}
}});

// Set this global variable to avoid problems with non-local Jupyter deployments
// Basically this line:
// https://github.com/AnalyticalGraphicsInc/cesium/blob/1.64/Source/Core/buildModuleUrl.js#L13
// Fails because of this:
// https://github.com/jupyter/notebook/blob/6.0.2/notebook/templates/page.html#L25-L27
// Also, finding out about CESIUM_BASE_URL was not very straightforward,
// see https://github.com/AnalyticalGraphicsInc/cesium/issues/8327
var CESIUM_BASE_URL = 'https://cesium.com/downloads/cesiumjs/releases/{cesium_version}/Build/Cesium/'

require(['cesium'], function (Cesium) {{
    var czml = {czml};

    var ion_token = '{ion_token}';
    if (ion_token !== '') {{
        Cesium.Ion.defaultAccessToken = ion_token;
    }}
    var viewer = new Cesium.Viewer('cesiumContainer-{container_id}', {{
        terrainProvider: {terrain},
        imageryProvider: {imagery},
        shouldAnimate : true
    }});

    // To have an inertial (ICRF) view
    function icrf(scene, time) {{
        var icrfToFixed = Cesium.Transforms.computeIcrfToFixedMatrix(time);
        if (Cesium.defined(icrfToFixed)) {{
            var camera = viewer.camera;
            var offset = Cesium.Cartesian3.clone(camera.position);
            var transform = Cesium.Matrix4.fromRotationTranslation(icrfToFixed);
            camera.lookAtTransform(transform, offset);
        }}
    }}
    // Temporarily disable inertial view
    // until we make it work with 2D Mercator view
    // and fix the zoom sensitivity, see
    // https://groups.google.com/d/msg/cesium-dev/vuXmepd4T2E/i71tq2I8EAAJ
    // viewer.scene.postUpdate.addEventListener(icrf);

    viewer.camera.flyHome(0);
    viewer.scene.globe.enableLighting = true;

    viewer.dataSources.add(Cesium.CzmlDataSource.load(czml));
}});
"""


@attr.s
class CZMLWidget:
    document = attr.ib(default=Document([Preamble()]))
    cesium_version = attr.ib(default="1.88")
    ion_token = attr.ib(default="")
    terrain = attr.ib(default=TERRAIN["Ellipsoid"])
    imagery = attr.ib(default=IMAGERY["OSM"])

    _container_id = attr.ib(factory=uuid4)

    def build_script(self):
        return SCRIPT_TPL.format(
            cesium_version=self.cesium_version,
            czml=self.document.dumps(),
            container_id=self._container_id,
            ion_token=self.ion_token,
            terrain=self.terrain,
            imagery=self.imagery,
        )

    def to_html(self, widget_height="400px"):
        return CESIUM_TPL.format(
            cesium_version=self.cesium_version,
            script=self.build_script(),
            container_id=self._container_id,
            widget_height=widget_height,
        )

    def _repr_html_(self):
        return self.to_html()



from io import StringIO

from czmlpy import Document, Packet


def test_document_has_expected_packets():
    packet0 = Packet(id="id_00")
    packet1 = Packet(id="id_01")

    document = Document([packet0, packet1])

    assert document.packets == [packet0, packet1]


def test_doc_repr():
    packet = Packet(id="id_00")
    expected_result = """[
    {
        "id": "id_00"
    }
]"""

    document = Document([packet])

    assert str(document) == expected_result


def test_doc_dumps():
    packet = Packet(id="id_00")
    expected_result = """[{"id": "id_00"}]"""

    document = Document([packet])

    assert document.dumps() == expected_result


def test_document_dump():
    expected_result = """[{"id": "id_00"}]"""
    packet = Packet(id="id_00")

    document = Document([packet])

    with StringIO() as fp:
        document.dump(fp)
        fp.seek(0)
        result = fp.read()

    assert result == expected_result
